@Comment{$Header: hostmanager.mss,v 1.2 88/12/02 16:16:39 jtkohl Exp $}
@Comment{$Source: /mit/zephyr/doc/techplan/RCS/hostmanager.mss,v $}
@Comment{$Author: jtkohl $}
@Comment{$Revision: 1.2 $}
@Comment{Copyright 1987, Massachusetts Institute of Technology}
@Part[hostmanager, Root "zephyr.mss"]

@Begin(Comment)
Random Editorial Comments ---------------------------------------------------

-----------------------------------------------------------------------------
@End(Comment)

@process(PageHeadings)

@Chap(The HostManager Client)
@Label(hostmanager)

@Section{Overview}

Each host in the network that supports Zephyr clients runs a
Zephyr HostManager Client.  When a host boots it starts a
HostManager which contacts a Zephyr server (designated by the Hesiod
name server), sending it a <ZEPHYR_CTL,HM,BOOT> notice specifying
that the server should flush all states previously associated with
this host.

@begin{comment}
This is no longer true.... inetd is not involved anymore.

  It waits for an acknowledgment, and exits upon receipt of
this acknowledgment.

A new HostManager is started by inetd the first time a Zephyr
notice is sent by a client program.  This HostManager will remain in
the background until it is killed at workstation deactivate time.
@end{comment}

The HostManager is the focus for all outgoing messages from the local
machine.  Messages are sent from client programs via Zephyr Client Library
function calls to the HostManager, which sends an acknowledgment to
the library routine.  It then relays the packet to the current
``owning'' server.  The server sends an acknowledgment back to the
HostManager.  The HostManager uses this to make sure that the server
is still responding.  It then forwards the acknowledgment to the client,
unless the notice type specified that the acknowledgment should not be
forwarded.

Whenever the HostManager fails to receive an acknowledgment of a
transmitted notice, it
retransmits the notice several times, after which time it seeks out a
new server.

The Zephyr server may occasionally send an <HM_CTL,SERVER,PING>
notice to the HostManager if it doesn't receive an acknowledgment
from a client program.  The HostManager simply sends back an
acknowledgment notice.

Upon receipt of a SIGHUP, the HostManager sends a <ZEPHYR_CTL,HM,FLUSH>
notice to the server, to indicate that all information about this host
should be flushed.  This can be used to insure that any stale
information in the server database is cleaned up when workstations are
no longer in use.  The HostManager then goes into the ``deactivated''
state.  In this state, it remains unattached to any server until a
client sends a notice to be forwarded to a server, at which point the
HostManager contacts a server, sends it a boot notice, and forwards the
client's notice.

@Section{HostManager Subsystems}

The HostManager is driven by a blocking read function call which waits
until a datagram arrives on its input port.  The packet is checked for
the host of origin.  If the host is the local loopback address (Internet
address 127.0.0.1), then the packet is sent to the transmission tower
routine, otherwise the server manager is called.

@begin{comment}
This doesn't exist yet... and it may never.
@SubSection{Local-host Manager}

The local-host manager routine is called whenever a notice arrives from
the local host.  If the notice is a registration message from a client,
that client is registered in a table for delivery of emergency messages.  If
it is a deregistration message, the client is removed from the table.
Otherwise, the transmission tower routine is called.
@end{comment}

@SubSection{Transmission Tower}

The transmission tower routine is called whenever a client program
sends a notice destined for a server.  If the HostManager is in
contact with a server, it forwards the notice to the server.  Whether
the HostManager sends the notice or not, it adds the notice to a queue
of unacknowledged notices, to be used should retransmission be
necessary.

If the notice is of kind HMCTL, the notice is handled internally by this
routine.  <HM_CTL,CLIENT,NEWSERV> notices cause the HostManager to call
the New Server routine.  <HM_CTL,CLIENT,FLUSH> notices cause the
HostManager to send a <ZEPHYR_CTL,HM,FLUSH> notice to its server, and go
into ``deactivated'' state.

@SubSection{Server Manager}

The server manager routine is called whenever a notice arrives from a
Zephyr server port on a machine other that the local host.  If the
originating host is not the current server machine and it is not an
acknowledgment (kind SERVACK or SERVNAK) or HMCTL (kind HMCTL) notice,
the receipt of the notice is logged and it is ignored.

The HostManager expects only three notice types from the server.
These are HMCTL control notices and packet reception acknowledgments,
which can be either SERVACK or SERVNAK types.

If an HMCTL notice is received, the notice opcode is examined.  If it is
``SHUTDOWN'', then the New Server routine is executed.  If the opcode is
``PING'', then the HostManager sends an HMACK notice back to the sender.
If the HostManager was not in contact with a server before the HMCTL
notice was received, the queue of unacknowledged notices is resent,
unless the notice received from the server is a SERVER_SHUTDOWN message.


If the packet received is an acknowledgment of a packet sent earlier,
the HostManager checks to see if the client wants the acknowledgment,
and if so sends the acknowledgment packet to the client.  It then
removes the packet from the queue of unacknowledged notices.

@SubSection{New Server Routine}

The new server routine is called when the server manager receives an
announcement that the server is going down, when a packet is not
acknowledged and the HostManager decides to change servers, or when a
client sends a new server notice to the HostManager.

The routine first sends an unacknowledged <ZEPHYR_CTL,HM,DETACH> notice
to the current server so that the server can remove its references to
this HostManager provided the server has not crashed, but is just
overloaded.  The HostManager then sends another Zephyr server a
<ZEPHYR_CTL,HM,ATTACH> notice.  When the new server acknowledges the
attach notice, it becomes the ``owning'' server.  The HostManager then
sends all of the unacknowledged packets to this new server, to insure
that all of them get delivered.

@SubSection{Queue Routines}

Whenever a client sends a packet destined for a server to the
HostManager, it is placed in a queue and then forwarded to the server.
The queue is a linked list, each element consisting of a timeout field,
a number of retries already sent, and the unacknowledged packet.  The
timeout field is set to the current time (in seconds) plus the number of
seconds to wait before a timeout should occur.

Whenever a timeout occurs, each entry in the queue is examined, and
the number of retries is incremented in each one which has a timeout
value less than the present time.  If the number of retries is greater
than the configured safety margin, the New Server routine is called.
If the number of retries is less than the safety margin, the
packet is resent to the server.

Whenever a new server has been successfully contacted, the entire
queue is resent.  Each entry in the queue is sent, its number of
retries is reset to zero, and the timeout field is set to the proper
value.

