@Comment{$Header: intro.mss,v 1.3 88/12/02 16:16:52 jtkohl Exp $}
@Comment{$Source: /mit/zephyr/doc/techplan/RCS/intro.mss,v $}
@Comment{$Author: jtkohl $}
@Comment{$Revision: 1.3 $}
@Comment{Copyright 1987, Massachusetts Institute of Technology}
@Part[intro, Root "zephyr.mss"]

@Begin(Comment)
Random Editorial Comments ---------------------------------------------------

-----------------------------------------------------------------------------
@End(Comment)

@Majorheading(Other documents of interest)
Besides this document, there are other documents describing parts of the
Zephyr system:
@begin(itemize)
@i(The Zephyr Programmer's Manual) gives a description of the Zephyr
Client Library, including examples of use.

@i(The Zephyr User's Manual) gives a user-level description of the
operation of the Zephyr Notification Service, including examples of
typical use.

@i(The Zephyr Installation and Operation Guide) describes typical
installation, operation and maintenance of the Zephyr system.

@i(Zephyr Implementation Notes) discuss the current implementation of
the Zephyr system.

@C(Unix) manual pages briefly describe the programs in the
Zephyr suite.
@end(itemize)

@Majorheading(Organization of this Document)
This document presents the concept of a notification service, including
design considerations, assumptions and constraints of Zephyr, and
details the design, constraints, and decision rationales for each of the
pieces of Zephyr.

@Majorheading(Definitions)
@begin(Description)
IP@\Internet Protocol.

UDP@\User Datagram Protocol.

Notice@\The fundamental unit of data transmission in the Zephyr system.
A notice consists of routing information in its @i(header) plus
client-determined @i(data).

ZID@\A ``Zephyr ID'' which uniquely identifies an entity using Zephyr.
In the Athena environment, a ZID is usually the Kerberos @cite(kerberos)
Principal identifier of the entity, such as
``jtkohl.root@@ATHENA.MIT.EDU.''

ASCII@\American Standard Code for Information Interchange.

ACL@\Access Control List.  An ACL lists in some form the entities
permitted to perform some action.

Class@\A part of the notice header specifying a general class of notice
type.  Often considered an attribute of a notice.

Instance@\A part of the notice header specifying a subclass of the
notice class.  Sometimes referred to as ``Class instance.''

Recipient@\A part of the notice header specifying the destination ZID of
the notice.  The recipient may be the null string ("\0") to indicate that
the notice should be multicast.

Multicast@\To transmit to a subset of all clients.

Zephyr Client@\Any program or user agent which can transmit or receive
notices.

Subscription@\A tuple of {class,instance,recipient} maintained by a
Zephyr server indicating a client's desire to receive notices with the
specified class, instance, and recipient fields in its header.

Wildcard@\A string specification that matches all strings for comparison
purposes.

Zephyr Realm@\The set of hosts supported by a particular set of Zephyr
servers.  This will typically be the same as a Kerberos realm.
@end(Description)  May also refer to the name given this realm.

@Chap{An Introduction To The Design Of Zephyr}
@Label(introduction)

This chapter is an introduction to the concept of a
notification service in general and to the design of the Athena 
Notification Service, Zephyr.  The sections which follow address
the following issues:

@Begin(Itemize)

What role does a notification service play in modern network based
workstation environments?

What are the motivating factors behind the development of a
notification service?

Under what design constraints must the service be developed?

What environmental assumptions have been made in the design and
implementation of Zephyr?

What level and types of services should users expect from a
notification service?

What are some of the unsolved problems and topics for future
development?

@End(Itemize)

While much of the information in this chapter is oriented
around Zephyr, Project Athena's Notification Service, it is our
belief that the concepts presented here can be generalized to fit a
broad range of notification services and systems.

@Section{Why a notification service?}


As Project Athena deployed large numbers of workstations and
servers, it became clear that current @c(Unix) communications systems were 
inadequate and incapable of dealing with many of the needs of the
environment being put in place at M.I.T.  The advantages of a
notification service as a solution to a broad spectrum of
communications needs in a workstation environment became clear as we
began to examine these inadequacies.

When services designed for use in a time-sharing environment
are reused for a very large system of networked workstations, certain
communication services begin to fail@Foot{Services, in general, begin
to fail, but the scope of this discussion is primarily the realm of
communication services.}.  Their failure is predominantly due to their
inability to cope with increasing network size (i.e., an increase in
the number of both workstations and local area networks).  In
examining how certain of these services communicate with their
clients, we have identified two primary failure modes: the
inability of a service to cope with rapidly increasing numbers of client
nodes, and the inability of clients to deal with the
replacement of a local service with a remote service.
Zephyr is the result of what was initially begun as the
development of a solution to these two failure modes.  Zephyr
has grown into a more powerful tool than was originally
anticipated; what began as the development of a ``desirable'' service
soon turned into the development of a ``required'' service.

The following lists some of the existing workstation
communication requirements that have been helped to scale properly by
relying on the existence of a notification service.

@Begin(
	Description, 
	Indent -2.0 inches, LeftMargin +2.0 inches,
)

File Service@\Users need to be informed of unexpected changes in file
server status.  Utilizing the notification service, a file server
can send notices to the users and hosts that it knows would be
affected by a change in file server status.  If a user registers
an appropriate subscription with the subscription service, he will
receive any such notices.  A third
party client would be able to notify subscribers
about operational issues associated with that file server.

Post Office Service@\Remote
post offices can notify users about the arrival of new mail.

Electronic Conferencing Systems@\Electronic conferencing systems do not
necessarily keep a readership list (conferences are often are open to the
``general public''), so the system cannot identify a community of
interest for event notification.  However, users interested in the
status of certain meetings can use the subscription
service to subscribe to conference update notices.

Print Service@\Print servers (and queuing services in general) can
utilize the notification service to communicate status
information back to the user who submitted a job, whether or not the
user has changed location in the interim.  Typical replies might be
``Print job complete.'' and ``Paper tray 2 empty on LPS40-1.''.

MOTD Service@\Message-of-the-day information (system wide, service
specific or local) can be sent to users via the notification service
when they begin using a particular service.  Notification of changes to
that information can be provided via the subscription service.

On-Line Consulting@\The notification service can be used as the
underpinning of a dynamic on-line consulting service.  The subscription
service can be used to provide topic based information routing, user
location, and consultant to client rendezvous.

Host Status Service@\Broadcast based host status systems (such as ruptime) do
not scale to a large workstation environment; disk usage grows
linearly with network size, total packet compute time grows
geometrically, and broadcast facilities are available only on the local network
(broadcasts do not propagate through network gateways).
The notification service can provide asynchronous and
immediate host status and error log notification on selected hosts or
servers.  Notification can be initiated either by the server using
the notification service to reach predetermined recipients, or
to self-selected recipients through the subscription
service.  This is desirable for service management purposes.

User Location Service@\Broadcast based user location systems (such as rwho)
also do not scale to a large workstation environment for the same
reasons noted above under Host Status Service.  The notification
service can provide asynchronous and immediate user location and state
change (login/logout) notification on selected users.
This can facilitate communication among users.  For example,
within the limits of user permission and access control (described
elsewhere), students could watch for their friends, Teaching Assistants, or
colleagues.

Talk or Phone Service@\Current systems that do not have access to
network-wide user location information require their users know the
exact address of someone is in order to establish a two-way
communication channel.  In addition, such systems typically do not
provide any access control or method of rejecting an annoying ``talker''
beyond turning off the service altogether.  Using the notification
service, a ``talk'' facility can be constructed that locates the party
being called, transmits a connection request notice to that party and,
if permission is granted, automatically (via rendezvous information
included in the request notice) establish a connection.

Emergency Notification@\There is a requirement to provide a simple,
asynchronous and secure means of sending urgent notices to all users
on a workstation or in a particular group of workstations.  Mail is
slow and synchronous (in that the user must initiate the reading
process).  Broadcast methods are not useful due to scale and network
boundary considerations.  In addition, the workstation user must trust
the broadcasting host and the person issuing the message.  Finally,
broadcast does not answer the question of how to notify multiple users
on a workstation, should that be the case.  Using the notification
service, emergency notices can be sent directly to all users on any
specified host.

Message Service@\Current one-way communication services suffer the same
problems 
noted above under Talk or Phone Services.  Using the notification
service the implementation of a ``write'' utility becomes extremely
easy; almost all the work is subsumed by the notification service.
Notices can go to individual users or to multicast subscription groups.

Other Service Events@\These are only a few of the interesting
services.
The notification service can be used to reliably notify users of a
whole range of asynchronous service events that occur in current
distributed workstation environments.
@End(Description)

@Section{Design Requirements And Constraints}

A notification service is intended to provide network based
services with immediate, reliable and rapid entity-based communication for
small quantities of time-sensitive information.  Such a service must be
able to efficiently provide these capabilities, with the highest
possible fan-out (i.e., client to server ratio) but without adversely
affecting network load or server host performance.

As the notification service is a distributed service
intended to serve a large number of workstations, it is important that
all aspects of service maintenance and operation be easily
controlled and manipulated remotely.

Since the notification service is still evolving, it is
important that it be designed such that there is some method of
gracefully handling protocol compatibility from one version of the
service to the next.  This means that servers and clients must be able
to interoperate with at least the previous and next versions.

The notification service is a communication system and
depends upon an underlying network transport mechanism, so it must accept
certain design constraints imposed by that mechanism, and should make
attempts to remedy deficiencies present in that mechanism.

In order to provide entity-based communications, the notification
service must maintain a database that maps entity names to the current
location(s) in the network of the corresponding entity.  The only
requirement on the entity name used in 
addressing is that it be unique across the entire network.

@Section{Environmental Assumptions}

The following items are assumed in the design of Zephyr:
@begin(Description)
Networked environment@\Zephyr is designed for use in
a networked environment of (marginally) cooperating hosts.  For a
stand-alone single-user or timesharing system, it offers very little
advantage over standard @c(Unix) utilities such as @i(write) and
@i(talk).

High-speed network@\Zephyr assumes the availability of a high-speed
network for communication between hosts within a given Zephyr realm.
If a high-speed network is not available, many of the implicit
assumptions are no longer correct and the system may not perform
adequately.  There is @i(no) requirement for broadcast capability.

Authentication@\Zephyr assumes the presence of an authentication system
which is secure on the network over which it operates.  Because of its
flexibility, Zephyr needs a way to reliably determine the sender of any
given notice.  Without an authentication system in place, the
reliability of this identification is highly suspect, and the ability to
spoof Zephyr becomes easy.  [If Zephyr is being operated in a captive,
friendly, cooperative environment, this may not be a cause for concern.]

Administration@\The Zephyr server is designed to run indefinitely without
intervention.

@end(Description)

@Section{Fitting The Tool To The Job}

To satisfy the above requirements, Zephyr is composed of three principal
elements:
@begin(itemize)
A notice transport mechanism.

A set of ``canned'' applications which use the transport mechanism.

A library of functions through which new applications may use the
transport mechanism (the ``Zephyr Client Library'').
@end(itemize)

The core of the notice transport mechanism is a system of dynamically
updated, locally authoritative servers that provide centralized routing,
queuing and dispatching.  Clients communicate with these servers via the
Zephyr Client Library interface.  The Zephyr Client Library implements
the Zephyr Protocol, a reliable, authenticated, ZID to ZID notice
transmission protocol.

A Zephyr server requires no maintenance for daily operations; changing
access control lists or default subscriptions can be accommodated by
remote logins to the server host or by automatic database-fed updates
(such as those provided by the Athena Service Management System @cite[sms]).

The Zephyr Protocol includes a version number so that multiple versions
may be supported simultaneously.

The Zephyr Protocol is UDP/IP based.  We chose UDP/IP for several
reasons.  The campus network at M.I.T@. uses IP almost exclusively, so
we were constrained to the IP protocol suite.  We also desired a
``lightweight'' transport mechanism with little overhead.  TCP/IP has
far too much overhead involved in initiating connections.  UDP/IP is
much more appropriate, since it is connectionless and thus has much less
overhead.  The 4.3BSD implementation of UDP/IP allows rapid switching of
destinations, a necessity if we are to use a central server to do notice
dispatching.

Since Zephyr is based on UDP/IP, it is constrained to operate within the
capabilities of UDP/IP.  However, there is nothing in the design of the
protocol that would prevent it from using other network transport
mechanisms (such as a remote procedure call system).

The five primary UDP/IP imposed constraints are listed below, along with
a brief description of how they may or may not be visible to Zephyr
application programmers and end users.

@Begin(
	Description, 
	Indent -2.0 inches, LeftMargin +2.0 inches,
)
Duplicate Notices@\UDP/IP does not provide any suppression of
duplicate packets.  This means that it is possible that Zephyr
clients may under certain adverse network conditions receive duplicate
Zephyr notices.  Zephyr applications must be capable of
dealing with this possibility even though Zephyr does provide some
filtering to block duplicate notices.

Missequenced Notices@\UDP/IP does not provide packet sequencing.
While Zephyr notices do contain timestamps it is up to the
application to check the timestamp and be willing to deal with notices
received out of sequence@foot{There is still the problem of
synchronization of clocks on the (possibly many) sending hosts, so even
the timestamps cannot be relied upon to determine the order in which
notices were transmitted.}.

Flow Control@\UDP/IP does not provide any flow control capability.
The application must be capable of dealing with notices at whatever
rate they arrive or be willing to lose notices.

Unreliable Delivery@\UDP/IP does not provide a reliable delivery
mechanism.  While Zephyr does provide retransmission and several levels of
acknowledgment processing, it is up to the
application to decide how much overhead it is willing to incur in order
to guarantee notice delivery.

Packet Size@\UDP/IP packets have a relatively small, fixed size.
Considering the amount of routing data and other information that Zephyr
must store in each packet, there are some strict constraints on how much
user data may be included with each packet.  To help remedy this, Zephyr
provides automatic packet fragmentation and reassembly, but it is not as
efficient as other means of transmitting large quantities of data.
Applications requiring large data transfers should use Zephyr as a
rendezvous service, and establish their own separate data transmission
channels.
@End(Description)

How visible these constraints are to the end user is up to the
Zephyr application programmer.  For example, the ``zwrite''
application (which allows workstation users to exchange ``write''-like
messages) only guarantees that the message was sent, not that it will
actually arrive or how many copies will arrive.  This is because
constraints one and four above (Duplicate Notices and Unreliable
Delivery) were by design not completely eradicated.

Zephyr manages a ZID to location translation database, known
as the ``user location database.''  This database maps ZID's to tuples
of location information: hostname, IP port number, and display device
(among other things).  This database is made available to Zephyr clients
via the ``user locator'' service.  The only ZID's normally found in the
database are those designating users of the system; by convention
network services do not announce themselves to the user location
database, as the database is intended for use by users, not services.
Zephyr maintains a separate location database (containing
only ZID, IP address and port number) for all clients wishing to receive
notifications.  It is this database which is used to determine the
proper routing of notices.

The reliability of the information stored in the user location
database imposes some constraints upon applications that rely on
Zephyr and the users of those applications.
@Begin(Itemize)
User location information stored in the database @P{can} be
assumed to be a reasonably accurate report of a user login.  This can
be asserted because all user logins reported to Zephyr must be
Kerberos authenticated.

User location information stored in the database @P{can not}
be assumed to accurately indicate that a user @P{has not} logged
out.  This must be asserted because there is no way to guarantee an
orderly authenticated user logout (@i(e.g.) a workstation may hang, crash,
or reboot due to problems beyond Zephyr's control).

User location information @P{not} stored in the database
@P{can not} be used to assume a user @P{did not} login.  This must be
asserted because a user may choose to not make his or her login
information available.
@End(Itemize)

As designed, Zephyr performs a reasonable amount of housekeeping to
prevent transient data stored in the user location database from
persisting when it is no longer valid.  If a workstation crashes, the
user login sessions on that workstation terminate without sending logout
notices to Zephyr.  This permits invalid information to remain in the
database (Note: the initial logins were not invalid; that the
information remains in the database is.).  In order to cope with this, a
Zephyr client runs when a workstation reboots, telling Zephyr to flush
any previous state information associated with the rebooted workstation.

If Zephyr application programmers and Zephyr application users
understand the above issues and maintain realistic expectations then the
few difficulties should arise.

Zephyr clients determine what level of service they require from the
notification service by choosing which type of routing they use.  For
example, certain client services have a concise knowledge of who their
clients are and only need the notification service to route information
to those clients, using their ZID (the most basic service).  A file
server that knows which users it is serving needs only the ability to
reliably notify those users about service state changes.  On the other
hand, client services that cannot identify their clients (or may simply
not know who is interested in such state information) may wish to notify
``interested parties'' about service state changes.  A workstation event
logger would fall into this category.  This type of service would make
use of the subscription service with multicasting based on who has
declared themselves ``interested.''  Such a service layer provides the
ability to store communication state information for client services
external to those services.  This adds to the notification service the
ability to provide to its clients status and availability
information about other client services even when those client services
are disabled and, as such, can't communicate with their own clients
directly.

@cb(We do not specify how this system of services should
be implemented.)  Our current implementation of Zephyr makes both
the user locator service and the subscription service
functions of a single base notification server.  While this
implementation may not be optimal due to too tight a coupling between
the individual services, it proved easiest to implement within a tight
schedule.  It is our intention, at some future date, to further
separate the individual services, thereby gaining the flexibility of a
system of modular and abstract services.

In Zephyr, notices are understood to consist of two parts, a routing
header and client data.  It is the job of the Zephyr servers to route
notices from Zephyr clients to other Zephyr clients and servers based
upon attributes specified in the notice's routing header.  Servers
should never expect to be able to examine a notice's client data; it is
entirely possible that that data is encrypted or otherwise
uninterpretable.  By examining the attributes in a notice's routing
header, it is possible for any Zephyr server to compute a finite list of
recipients to whom the notice should be sent.  The most basic routing
attribute that may be specified is a tuple of <class, instance,
recipient ZID>.  Additional and more complex routing attributes might
need to be specified for layered services that use more complicated
routing methods.  For example, an experimental rule based routing
service might need specialized keywords.

We refer to the process of determining multiple notice recipients based
upon routing header attributes as ``multicasting.''  Multicasting is a
passive routing technique; attributes not recognized by a routing
service are simply ignored.  This allows layered services to implement
different notice routing methods that peacefully coexist while
utilizing the same base notification service.  It is not our intention
that scores of notice routing methods be immediately available, but
rather that the notification service fulfill an immediate need without
blocking future development.  Multicasting differs from other common
message routing techniques.  It is more efficient and less subject to
failure due to diseconomies of scale than broadcast techniques because
the complete set of recipients for any notice can always be determined,
and uninterested recipients need not process a notice only to discover
it should be ignored.  It requires less maintenance and incurs less
administrative overhead than traditional list based message transmission
techniques (such as electronic mailing lists) because additional
resources, routing methods and recipients (within limits) may be
dynamically added by almost any user.  Multicasting is a notice routing
technique that is well suited for use as the core of a notification
service.

A good understanding of Zephyr in perspective can best
be acquired by comparing the Zephyr Notification Service and a
more traditional method of workstation message delivery, electronic
mail.  The following table compares some of the critical metrics of
notice/message delivery that are applicable to both Zephyr and
electronic mail delivery by a typical @c(Unix) mail delivery system,
@i(sendmail).

@Tag(comp)
@Define(TableText=Text, Indent 0, Size 10)
@StandardTable(
	Name Comparison,
	Columns 3,
	FirstColumn FlushLeft,
	OtherColumns TableText,
	OtherColumnWidth 2.25 inches,
	Boxed, FlushTop
)
@Begin(Comparison, Above 2 lines, Below 2 lines)
@TableID(Comparison)
@TableHeading(
	Immediate,
	RowFormat ComparisonTopHeading,
	Line 'A Comparison Between Zephyr And Mail'
)
@TableHeading(
	Continued,
	RowFormat ComparisonTopHeading,
	Line 'A Comparison Between Zephyr And Mail (continued)'
)
@TableHeading(
	Immediate, Continued,
	RowFormat ComparisonColumnHeadings,
	Line 'Metric@\Zephyr@\Electronic Mail (Sendmail)'
)

Addressing@\Implicit/Dynamic: All addressing is dynamically
determined; an explicit ``address'' is not required.  One-to-one
addressing is supported via explicit specification of recipient
ZID.  In addition, subscription based
notification allows the recipient to be determined by notice
attribute information.
@\Explicit/Static: Sender must know and
explicitly provide the name and address (except for ``local'' mail) of
each recipient or list of recipients being sent to.  Static mailing
list support is provided.  Only recipients explicitly named will be
delivered to.
@Hinge()

Delivery Method@\Notices are delivered via dynamically routed
datagram.  No connections need be established or maintained.  Multiple
levels of notice acknowledgment are supported, to provide reliability as
needed.
@\Mail is delivered
via point to point SMTP connection.  Acknowledgments are not
supported per-se, but return receipts may be requested (but may not
work in all SMTP implementations).
@Hinge()

Delivery Action@\Asynchronous/Active: Notices arrive and are displayed
without user intervention.
@\Synchronous/Passive: Mail is read
by user action.  Mail, in general, is delivered to one
particular place (a ``post office'' or ``mail drop'') for each user; he
or she must then actively retrieve it.
@Hinge()

Message Length@\Usually short notice length, with support for
fragmentation of larger messages.
@\Long, typically unfixed,
message length. Mail messages may be and often are extremely large, on
the order of many pages of text.  When message length is fixed it is usually
done so by unpredictable rules that vary from site to site.
@Hinge()

Message Persistence@\Notices are considered time sensitive, no queuing
is provided.  If the destination client is not on the network then the
notice is discarded.  If a user didn't see it, he probably didn't need
to.
@\Long time-to-live.  Messages typically remain in your mail drop
until you retrieve and delete them.  Two weeks' worth of junk mail
remains two weeks' worth of junk mail.
@Hinge()

Message Fan-out@\High fan-out: Sending to large lists is efficient.
When multicasting, each client sends only one copy of a notice
regardless of the number of recipients.  Each server receives only one
copy of a notice being routed regardless of the number of recipients.
Client-determined lists require separate notices for each recipient.  No
notice queuing is provided by the server.  Clients determine whether or
not to retain notices.
@\Low fan-out: Sending to large lists can be @P{very}
resource consuming.  If a message is sent to many users many copies
are generated by the mail processing system, each of which is retained
until deleted by its recipient.  Queuing due to inability to contact
destination hosts can consume large amounts of file storage space.
@Hinge()

Traffic Performance@\High volume/High throughput: Notices
may be transmitted in large numbers due to the low overhead of
dynamically routed datagrams.
@\Medium volume/Low throughput: Multiple large mail messages can send a
reasonable volume of data but slowly as connections need to be
established and routes determined.
@Hinge()

System Configurability@\Dynamically reconfigurable: Dynamic resource
allocation and configuration within the base notification services
allows automatic and simple user level reconfiguration of other
services.
@\Statically reconfigurable: Reconfiguration of @i(Sendmail) is wizard
level work and has significant global impact.  No utilities are provided
for dynamic system modification or reconfiguration.  All changes must be
made centrally and atomically.
@Hinge()

System Maintenance@\Low maintenance: Zephyr servers dynamically recover
unused resources through time-outs and reference counting.  <class,
instance, ZID> tuples, ZID location information and other subscription
service resources have a well defined time-to-live.
@\High maintenance: Mail requires a post office staff to maintain post
office boxes (mail drops), mailing lists, the routing system and to
manually reroute undeliverable ``dead letters.''

@End(Comparison)

@Section{Future Directions And Unsolved Problems}

Once the basic notification service is in place it becomes a
simple matter to provide many other layered services based upon it.
The ``talk'' service mentioned above is a good example of a service
that utilizes multiple Zephyr services: the user locator service and
the notice routing service.

We envision Zephyr as a transport service that can incorporate new
notice routing methods as they are developed.  Because of the dynamic
configurability of the subscription service, Zephyr allows communication
development efforts to occur side-by-side with running production
systems that utilize Zephyr.  For example, we are currently working with
researchers at MIT's Sloan School of Management who are looking at using
Zephyr as the transport service layer for a rule based communication
system @Cite(wendy).  Such a system under development could use the
subscription service to gather and process notices, and when complete
could coexist with the subscription service as an alternate routing
method.

Following is a list of some of the areas that the authors have slated
for future development.  They are either unsolved problems or areas that
we feel need further investigation.

@CBon()
@Begin(Itemize)

Make the Zephyr Protocol and retransmission algorithms robust enough for
use across long-haul and/or lossy networks.

Modify the Zephyr server to interact sensibly with other Kerberos
authentication realms, including problems of user registration across
realms and notice forwarding across realms.

Develop a more formal interface definition for use between the
Zephyr notice transport layer and Zephyr routing services.

Develop a more advanced user interface for the Zephyr WindowGram
Client.

@End(Itemize)
@CBoff()
