@Comment{$Header: server.mss,v 1.3 88/12/02 16:17:27 jtkohl Exp $}
@Comment{$Source: /mit/zephyr/doc/techplan/RCS/server.mss,v $}
@Comment{$Author: jtkohl $}
@Comment{$Revision: 1.3 $}
@Comment{Copyright 1987, Massachusetts Institute of Technology}
@Part[server, Root "zephyr.mss"]

@Begin(Comment)
Random Editorial Comments ---------------------------------------------------

	Auto-justified mode is disgusting.  Especially for scribe mode.

-----------------------------------------------------------------------------
@End(Comment)
@process(PageHeadings)

@Chap(The Zephyr Server)
@Label(server)

The Zephyr server manages subscriptions, default subscriptions,
user location information, and access control lists, and redistributes
notices between clients.

@section(Server Authority)
Each server is authoritative only for those client hosts which have
associated themselves with it.  All other data the server holds is
considered correct but may be invalidated at any time by command of the
server which is authoritative.  This scheme provides a distributed
database between the servers.
Each server may correctly redistribute any notice without
consulting any other server.  Incremental update is available, but
without the complexity of maintaining complete consistency between all
the servers, at the cost of associating with each piece of information
an identification of the server which controls that information.

For efficiency, we desire that each server maintain the entire
database of subscriptions and locations.  But maintaining a fully
distributed database with incremental update is too difficult to be
worthwhile for Zephyr.  So we have compromised our desire for
distribution with the reality of maintenance and implementation
constraints.

@Section(Server Initialization and restart)
When a server is started, it either asks the name service @i(Hesiod) @cite[hesiod]
or consults a file@foot(This choice is dependent on compile-time flags.)
to determine which other servers it should communicate with.  It puts
each other server into the DEAD state (see Figure@ @Ref(fsm)), schedules
an immediate HELLO packet transmission for each other server, and
commences operation.  In addition to the ``normal'' servers, each server
maintains the state of a ``limbo'' server.  This nonexistent server is
used for storing the state of hosts which were formerly associated with
a server which is now considered DEAD.  If a server loses contact with
another server for a significant amount of time, it considers that
server DEAD, and transfers the information which was formerly controlled
by the other server to ``limbo.''

The server also initializes the access control lists of any registered
classes in the class registry, and reads and caches the system default
subscriptions from a file.

@begin(FullPageFigure)
@caption<The Finite State Machine (FSM) of server states>
@tag(fsm)
@begin(transparent, fill, spread 0, above 0, below 0, Font Headingfont)
Each server maintains a separate FSM for each other known server.  There
are four major states: UP, TARDY, DEAD, and STARTING.

Transitions labeled with `e' are empty transitions which always occur,
requiring no events or input.

Boxes attached to states indicate actions performed upon entering that
state.

Timers are named T-xxx where xxx is a name representing the
timer's use.  Resetting a timer means resetting the expiration of the
timer to the full value.  T-xxx labeling a transition means that
transition is taken when the timer expires.
@end(transparent)

@Picture(Size=7 inches, ScaleableLaser="fsm.PS")

@end(FullPageFigure)

If two servers establish contact with each other, they exchange
authentication information (to verify identities) and any
authoritative data they hold and mark each other as UP.  They may also exchange
``limbo'' information if one of the servers does not have any hosts in
``limbo.''  This allows the information associated with each host to be
retained across server failures and restarts.

When exchanging information, a server may ask another server to send its
idea of the server's state and use that information as its authoritative
information.  This prevents loss of information if a server crashes and
restarts before other servers notice its failure.  If the failed and
restarted server did not retrieve its state from another server, any
client requests dependent on that lost state might be improperly
dispatched or discarded.

The servers normally communicate using the standard Zephyr
protocol.  If a server receives any response from another server it
previously thought DEAD, it opens a ``brain-dump'' socket and requests
the other server to connect and initiate the ``brain-dump'' protocol.
When communicating via the ``brain-dump'' protocol, servers use the
Zephyr library routines for packing and unpacking notices, but
instead of using the standard datagram transport they use a TCP/IP
connection for reliable transport for the ``brain-dump.''  If this
``brain-dump'' succeeds, each server schedules an immediate HELLO packet
transmission to its peer, which when acknowledged will result in a state
transition to UP.

@Section(Server Operation)
The Zephyr server operates in a dispatch loop.  Through its
datagram socket it receives notices from clients and servers.  These
notices are dispatched via appropriate server routines.  The server also
processes previously set timeouts, such as a retransmission timer or a
server contact timer.

@subsection(Timers)
The Zephyr server associates a timer with each other server.  The timeout of
this timer and the action taken when it expires varies depending on
the state of the other server in the current server's FSM.
Figure@ @ref(fsm) shows the FSM and the actions taken.

Each notice which has been distributed to a client but not acknowledged
has a timer which controls retransmission attempts.

Each inter-server update notice which has been forwarded to the other
servers but not acknowledged has a timer which controls retransmissions.

@subsection(HostManager interactions)
@label(failure)
Any notices which have been transmitted to clients but not acknowledged
are queued and retransmitted several times with a short interval between
retransmission attempts.  If the notice remains unacknowledged, the server
attempts to establish contact with the HostManager on the client's host.
If the server is authoritative for that host, it sends a HostManager
Ping packet to the hostmanager.  If the server is not authoritative, it
requests the authoritative server to send the ping.
If the server believes the host to be in ``limbo,'' the recovery attempt
is aborted.

If the ping is acknowledged, the state associated with the
unresponsive client is flushed.  If the ping is not acknowledged after a
short interval, the state associated with the entire host is flushed,
under the assumption that the machine has crashed@foot(Note that if only
the HostManager was dead, and not the whole machine, the clients on that
host would still have no way to transmit information via Zephyr,
and the flush is still appropriate.
See Chapter @ref(hostmanager) for more details on the HostManager.).

All notices originating from a HostManager are acknowledged.  The
acknowledgment may take one of several forms, depending on the
disposition of the notice.  If the notice was unauthentic or unauthorized and
attempted a function requiring authentication and/or authorization, the
acknowledgment is type SERVNAK
with message LOST.  If the packet assumed some internal
server state and that state was not present, the acknowledgment is type
SERVNAK with message FAIL.  Otherwise the type is SERVACK, and the
message is SENT if the notice was redistributed or processed without
error or LOST if no clients were subscribed to that notice.

@subsection(Notice handling)
When datagram notices are received, the server invokes the appropriate
dispatch routine to handle the notice, based on its origin and
the contents of its header.  The following table details the dispatch
rules.

@StandardTable(
Name displatch,
Columns 3,
AllColumns FlushLeft,
FlushTop,
HeadingBoxed,
ColumnLined)
@begin(displatch)
@tableid(displatch)
@tableHeading(immediate, RowFormat displatchTopHeading,
line 'Dispatch rules')
@tableHeading(continued, RowFormat displatchTopHeading,
line 'Dispatch rules, continued')
@tableheading(immediate, continued, RowFormat displatchColumnHeadings,
line 'Notice origin@\Notice class@\Dispatch routine')

A known server@ @ @ @ @ @\(any class)@\@t[server_dispatch]

Some other origin@\HM_CTL@\@t[hostm_dispatch]

@\ZEPHYR_CTL@\@t[control_dispatch]

@\LOGIN@\@t[ulogin_dispatch]

@\USER_LOCATE@\@t[ulocate_dispatch]

@\ZEPHYR_ADMIN@\@t[server_adispatch]

@\<anything else>@\@t[sendit]
@end(displatch)

When the notice originated from another server known to this server,
@f[server_dispatch] performs more dispatch processing:
@begin(itemize, above 0)
If the notice is an acknowledgment of some previously transmitted
notice (i.e. the notice has z_kind == SERVACK), the packet it
acknowledges is removed from the retransmit queue and the server is
marked as having responded (this may modify the state of the other
server and cause a notice transmission, according to Figure@ @ref(fsm)).

If the notice is of class ZEPHYR_CTL, LOGIN, or USER_LOCATE, it is
passed on to the corresponding dispatch routine, after determining the
true origin of the notice.  This case occurs when a server has acted on
a client request, and asks the other servers to update their databases
to maintain synchronization by transmitting the notice it received from
the client to the other servers.

If the notice is of class ZEPHYR_ADMIN, it is passed on to
@f[admin_dispatch], which dispatches on opcode:
@begin(itemize, above 0)
If the notice has opcode HELLO, the notice is a hello packet, and the
FSM for the originating server is updated appropriately.

If the notice has opcode IHEARDYOU, the notice is an acknowledgment of
a hello packet, and the FSM for the originating server is updated
appropriately.

If the notice has opcode GOODBYE, the notice is a shutdown packet,
indicating impending shutdown of the peer server.  All the state
associated with that server's hosts is put into the ``limbo'' state until
those hosts contact a new server, and the server is marked as DEAD in
the FSM.

If the notice has opcode DUMP_AVAIL, the server initiates a ``brain-dump''
connection to the peer server (see Section @ref(bdumpproto)).

If the notice has opcode LOST_CLIENT, the server initiates the
HostManager recovery protocol with the host of the client named in the
body of the message.  The body contains two strings, each
NUL-terminated.  The first is an ASCII representation of the address of
the client in Internet ASCII dot-notation.  The second is an ASCII
representation of the port number of the client.

If the notice has opcode KILL_CLIENT, the server flushes the state
associated with the client named in the body of the message.  The body
is in the same format as the LOST_CLIENT notice.

If the notice has any other opcode, its reception is logged and it is
ignored.
@end(itemize)
@end(itemize)

@f[hostm_dispatch] does further dispatch work, based on the opcode
of the notice:
@begin(itemize, above 0)
If the notice has opcode BOOT, the server flushes any state associated with
clients on the originating host, and adds (or replaces) the host on its
list of ``owned'' hosts for which it is authoritative, informing other
servers that it has assumed ``ownership'' of the host.  The other servers
mark the host as owned by the server which has claimed ownership.

If the notice has opcode FLUSH, the server flushes any state associated
with the originating host, but maintains the owning server of that host.

If the notice has opcode ATTACH, the server claims ``ownership'' of the host
and informs other servers that it has done so.  The other servers
mark the host as owned by the server which has claimed ownership.

If the notice is an acknowledgment of a server query resulting from a
failed client (see Section @ref(failure)), the server flushes the
information stored for the failed client, and turns off the timer set
when the query was sent.
@end(itemize)

@f[control_dispatch] does further dispatch work, based on the opcode and
instance of the notice:
@begin(itemize, above 0)
If the instance is HM, the notice is passed to @f[hostm_dispatch] for
processing.

If the opcode is GIMME, the originating client is returned a list of his
subscriptions.

If the opcode is GIMMEDEFS, the originating client is returned a list of
the system default subscriptions.

If the opcode is SUBSCRIBE, the originating client is subscribed to the
<class, instance, recipient> triples specified in the message body.
The message body contains null-terminated strings.
Each set of three is considered a triple, starting at the beginning of
the message body.  If the client had no previous subscriptions, the
client is also subscribed to the system default subscriptions.

If the opcode is UNSUBSCRIBE, the originating client is unsubscribed to
the triples specified in the message body.  The format of the body is
identical to that of the SUBSCRIBE operation.

If the opcode is CLEARSUB, the originating client is unsubscribed to all
the triples he is currently subscribed to.

If the notice has any other opcode, its reception is logged and it is
ignored.
@end(itemize)
@f[ulogin_dispatch] does further dispatch work, based on the opcode of
the notice:
@begin(itemize, above 0)

@begin(multiple)
If the opcode is OPSTAFF, REALM-VISIBLE, REALM-ANNOUNCED, NET-VISIBLE,
or NET-ANNOUNCED, a location for the ZID of the originating client is
added to the location database, with the exposure set as the opcode
directs, and other fields filled in as specified in the body of the
notice.  The body contains three NUL-terminated strings: the machine
name of the client's host, the time of the exposure change, and the name
of the display device in use by the client.

If the exposure is sufficiently broad, a login notice is fabricated for
this user (with opcode set to USER_LOGIN) and transmitted to those
clients which are both permitted to receive such notices and subscribed
to login messages for that ZID.
@end(multiple)

If the opcode is NONE, the location (as specified in the notice) of the
ZID of the originating client is removed from the location database.

If the opcode is USER_LOGOUT, the ZID of the originating client is
removed from the location database, and if his exposure is sufficiently
broad, the notice is retransmitted to those clients which are both
permitted to receive such notices and subscribed to login messages for
that ZID.

If the opcode is USER_FLUSH, all locations for the ZID of the
originating client are removed from the location database.

If the notice has any other opcode, its reception is logged and it is
ignored.
@end(itemize)
@f[ulocate_dispatch] does further dispatch work, based on the opcode of
the notice:
@begin(itemize, above 0)
If the opcode is LOCATE, any locations of the ZID named in the instance
of the notice which the sender is authorized to receive are sent to the
port specified in the port field of the notice.  Each location is
returned in the message body as a triple {machine_name, login_time,
terminal} of NUL-terminated strings.  Successive locations are appended
to the message body.

If the notice has any other opcode, its reception is logged and it is
ignored.
@end(itemize)
@f[server_adispatch] does further dispatch work, based on the opcode of
the notice:
@begin(itemize, above 0)
If the opcode is STATUS, a notice containing server status information
in the message body is returned as an acknowledgment to the originating
client.

If the notice has any other opcode, its reception is logged and it is
ignored.
@end(itemize)

When the notice is some other class, the notice is passed to @f[sendit],
which checks authorization for transmission of this class, and then
transmits the notice to all clients subscribing to it.
Most notices will be of this type.

@subsection(Subscriptions)
The Zephyr server maintains a list of subscriptions for each
client.  Each subscription contains the class, class instance (possibly
a wildcard), and recipient of the notices requested.  @i(For security
considerations, the recipient may only be either the ZID
of the subscribing client or a null recipient.)  A null recipient is
effectively a multicast address.  Any interested client may
subscribe to a class/instance pair with a null recipient field (subject
t access control restrictions on that class),
and any notice transmitted to the same class/instance pair with a
null recipient field will be redistributed to all those subscribed
clients.

In addition to the subscriptions explicitly requested by a client (via a
ZEPHYR_CTL/SUBSCRIBE notice), all clients are automatically subscribed
to a set of default subscriptions (the list of default subscriptions is
stored in a file on the server host.).  A client may unsubscribe to any
of these defaults via the standard ZEPHYR_CTL/UNSUBSCRIBE mechanism.

@subsection(Classes and Access Control) 
There are two types of classes maintained by the server, registered and
unregistered classes.  Any client may send or subscribe to unregistered
class notices.
The class registry is a file specifying which classes
are registered.  Each registered class is associated with four access
control lists.  Each list specifies the ZID's with the
authorization for the function associated with the list.  It is
permissible to place wildcards in the access control lists.
The four authorization types are:
@begin(itemize)
SUBSCRIBE authorization.  If a ZID is named in this list, it may
subscribe to notices in the class.  Otherwise, it may not subscribe to
notices in this class.

INSTWILD authorization.  If a ZID is named in this list, it may
specify a wildcard instance in a subscription to notices in the class.
Otherwise, it may only subscribe to non-wildcard instances of the
class.

TRANSMIT authorization.  If a ZID is named in this list, it may
transmit notices of the class.  Otherwise, it may not transmit notices
of this class.

INSTUID authorization.  If a ZID is named in this list, it may transmit
notices of this class with any instance.  Otherwise, it may only
transmit notices of this class with instance equal to its own ZID.
@end(itemize)

If the file representing the access control list specifying an
authorization type for a given registered class is missing, all
authenticated ZID's are granted that authorization, but no
unauthenticated ZID's are granted authorization.

@subsection(Instances)
The server does not make any restrictions on instances except those
imposed by access control lists.

@subsection(User Locations)
The Zephyr server maintains a table of user locations.  Each entry
includes the ZID of the client at the location, machine name, time of
login, display device name, and an exposure level.  All elements but the
ZID are stored exactly as reported by the login notice.  The ZID is
extracted from the authentication information.

@Section(Server-to-server Download Protocol)
@label(bdumpproto)
When two servers establish contact, they initiate the Server-to-server
Download protocol (Brain-dump protocol).
If at any time during the brain-dump exchange an error occurs, the
entire dump is aborted.

When the brain-dump is completed, each server will have nearly identical
copies of the database information (the information assigned to ``limbo''
and to other servers not a party to the brain-dump operation may
differ).  In effect, the brain-dump re-plays all the operations
necessary to establish the state present in the servers.

The initiating server sends its peer an opcode ``DUMP_AVAIL'' notice
(described below in Section @ref(srvclass)).  When a server receives
such a notice from another server, it attempts to connect to the
indicated TCP port.  Upon success, it sends a Kerberos ticket and
authenticator to the peer, and waits for a ticket and authenticator in
return.

When a server which is listening to a TCP socket receives a connection
request, it accepts the connection and closes the listening socket.  It
then reads an authenticator from the newly accepted socket, and if
authentication yields the principal ``zephyr.zephyr'' in the server's
Kerberos realm, it forms an authenticator and sends it to the
peer@foot(This is a crude form of mutual authentication; in the future
a more secure form of mutual authentication involving decrypting,
modifying, and returning random data will be used.).

After authentication succeeds, the peers begin the actual data
transfer.  The server which bound and listened to a socket begins a
send loop, and its peer begins a receive loop.  When this transaction is
complete, the roles are reversed.

The data transmission is accomplished by using the standard Zephyr
library routines to format and decompose notices into and from a
character stream. Each formatted notice is transmitted via the TCP
connection to the peer, preceded by a two-byte length field (in
network byte-order) specifying the length in bytes of the following
packet.

The server in the receive loop (the receiver) asks the sender for data
in separate ``chunks.''  The receiver may, at its option, ask for the
``LIMBO'' data and ``MY_STATE'' data (data on hosts assigned to
``limbo'' and those assigned to the receiver), in that order. It then
must ask for ``YOUR_STATE'' (data on hosts assigned to the sender).

Servers which have no hosts in the limbo state will request ``LIMBO''
data.  In this way, hosts which have not been heard from since a server
crash-restart cycle are retained in the database.
Servers which have never received any information from another server
will request ``MY_STATE'' data.  In this way, if a server restarts quickly,
before a peer notices his absence, his authority is retained for those
hosts which have not yet chosen a new server.

For each chunk request, the receiver enters a loop.  Inside this loop it
expects one of the following:
@begin(itemize)
A boot notice for a host.

A login notice for a user.  The receiver must have received a host boot
notice before it will accept a login notice.

A client registration notice (Opcode ``NEXT_CLIENT''). The receiver
must have received a host boot notice before it will accept a client
registration notice.

A subscription notice.  The receiver must have received a client
registration notice before it will accept a subscription notice.

An end of dump notice (Opcode ``DUMP_DONE'').
@end(itemize)

When an opcode ``DUMP_DONE'' notice is received, the loop is exited, and
the next chunk is requested.  If the receiver desires no more ``chunks'', it
is finished, and the roles reverse (if appropriate).

The server in the send loop waits for a request for a chunk.  It then
iterates over all hosts associated with the chunk requested, sending a
host boot notice, then all the locations on that host, and then sends
client registration and subscription notices for each client on that
host.  When finished with the set of hosts, it sends an opcode
``DUMP_DONE'' notice.

@subsection(Server brain-dump notice definitions)
Following is a table describing notices used in the brain-dump protocol
which are not part of the reserved class definitions (Section
@ref(resclass)).  All classes, instances, and opcodes used are NUL-terminated
ASCII strings.

@label(srvclass)
@Begin(Center, TableColumns BodyRowFmt, Use Box)
@TableID(BdumpClasses)
@TableHeading(
	Immediate,
	RowFormat TableHeadFmt,
	Line="Messages used for Brain Dumps"
)
@TableHeading(
	Continued,
	RowFormat TableHeadFmt,
	Line="Messages used for Brain Dumps (continued)"
)
@TableHeading(
	Immediate, Continued,
	RowFormat ColumnHeadFmt,
	Line="Class@\Instance@\Opcode@\Effect"
)

ZEPHYR_ADMIN@\<VERSION>@\DUMP_AVAIL@\Inform the other server that a
brain-dump is available.  The instance is the version number identifying
the brain-dump protocol.  The sender is the hostname of the originating
server.  The client data area contains two NUL-terminated strings: the
Internet address of the originating server, in Internet ASCII dot
notation, and the TCP port number from which the brain-dump is to be
obtained, in ASCII.

ZEPHYR_ADMIN@\LIMBO@\DUMP_AVAIL@\Request the other server to send the
state associated with hosts in ``LIMBO.''

@\YOUR_STATE@\DUMP_AVAIL@\Request the other server to send the state
associated with hosts assigned to the other server.

@\MY_STATE@\DUMP_AVAIL@\Request the other server to send the state
associated with hosts assigned to this server.

@\``'' (empty string)@\DUMP_DONE@\Indicate that this phase of the
brain-dump (or the entire brain-dump) is complete.

@\CBLOCK@\NEXT_CLIENT@\Begin a subscription transfer for a new client.
The sender field is the client's ZID.  The client data area contains two
NUL-terminated strings: The client's port number, in ASCII, and the
client's DES session key, in ASCII.  The client's Internet address is
assumed to be the address of the most-recently registered host.

@end(Center)
